package fr.numeco.misis.sitemap.service.impl;

import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.parsers.DocumentBuilder;

import org.eclipse.microprofile.config.inject.ConfigProperty;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import fr.numeco.misis.enums.AnalyseStatut;
import fr.numeco.misis.enums.SitemapMessage;
import fr.numeco.misis.sitemap.dto.SitemapDto;
import fr.numeco.misis.sitemap.exception.SitemapUrlException;
import fr.numeco.misis.sitemap.service.SitemapService;
import fr.numeco.misis.suividesite.domain.GroupeDePages;
import fr.numeco.misis.suividesite.domain.Page;
import fr.numeco.misis.suividesite.dto.PageFormDto;
import fr.numeco.misis.suividesite.repository.GroupeDePagesRepository;
import fr.numeco.misis.suividesite.repository.PageRepository;
import io.quarkus.logging.Log;
import io.smallrye.mutiny.Multi;
import io.smallrye.mutiny.infrastructure.Infrastructure;
import jakarta.enterprise.context.ApplicationScoped;
import jakarta.inject.Inject;
import jakarta.transaction.Transactional;

@ApplicationScoped
@Transactional
public class DefaultSitemapService implements SitemapService {

    public static final String PAGE_ATTRIBUT_URL = "url";
    private static final String SITEMAP_URL_REGEX = "^.*\\.xml(?:/??\\?.+)?";
    private static final Pattern SITEMAP_URL_PARTTERN = Pattern.compile(SITEMAP_URL_REGEX, Pattern.CASE_INSENSITIVE);

    @Inject
    private DocumentBuilder dBuilder;
    @Inject
    private PageRepository pageRepository;
    @Inject
    private GroupeDePagesRepository groupeDePagesRepository;
    @ConfigProperty(name = "sitemap.size.limit")
    private Integer sizeLimit;

    public void setGroupeDePagesRepository(final GroupeDePagesRepository groupeDePagesRepository) {
        this.groupeDePagesRepository = groupeDePagesRepository;
    }

    public void setPageRepository(final PageRepository pageRepository) {
        this.pageRepository = pageRepository;
    }

    @Override
    public SitemapDto readSitemap(final String sitemapUrl) {
        return this.readSitemap(sitemapUrl, false);
    }   

    @Override
    public SitemapDto readSitemap(final String sitemapUrl, final boolean requireFullResult) {
        final Set<String> urlsExtractedFromSitemap = new LinkedHashSet<>();
        Set<SitemapMessage> messages = new LinkedHashSet<>();

        if (sitemapUrl == null) {
            throw new SitemapUrlException("XML file is missing.");
        }

        final Matcher xmlMatcher = SITEMAP_URL_PARTTERN.matcher(sitemapUrl);
        if (!xmlMatcher.find()) {
            throw new SitemapUrlException("Not a XML file.");
        }

        try {
            final Document parsedSitemap = this.dBuilder.parse(sitemapUrl);
            parsedSitemap.getDocumentElement().normalize();

            final NodeList urlNodeList = parsedSitemap.getElementsByTagName("loc");
            final int nodeLength = urlNodeList.getLength();
            
            if (nodeLength == 0) {
                messages.add(SitemapMessage.EMPTY_SITEMAP);
            }

            Log.infof("%s URLs in sitemap %s", nodeLength, sitemapUrl);

            for (int i = 0; i < nodeLength; i++) {
                if (urlsExtractedFromSitemap.size() == this.sizeLimit) {
                    messages.add(SitemapMessage.NUMBER_OF_ITEMS_LIMIT_REACHED);
                    if (!requireFullResult) {
                        return new SitemapDto(urlsExtractedFromSitemap, messages);
                    }
                }
                final Node node = urlNodeList.item(i);

                if (node.getNodeType() == Node.ELEMENT_NODE && node instanceof Element urlElement) {
                    final String url = urlElement.getTextContent();

                    if (this.isValidUrl(url)) {
                        urlsExtractedFromSitemap.add(url);
                    } else {
                        Log.warnf("invalide url in sitemap %s", url);
                    }
                }
            }
        } catch (SAXException e) {
            messages.add(SitemapMessage.SITEMAP_FORMAT_ERROR);
        } catch (IOException e) {
            messages.add(SitemapMessage.SERVICE_UNAVAILABLE);
        }

        return new SitemapDto(urlsExtractedFromSitemap, messages);
    }

    @Override
    public Multi<Page> process() {
        final List<GroupeDePages> groupeDePagesSitemap = groupeDePagesRepository.findSitemapGroups();

        return Multi.createFrom().items(groupeDePagesSitemap.stream())
                .emitOn(Infrastructure.getDefaultExecutor())
                .runSubscriptionOn(Infrastructure.getDefaultExecutor())
                .onItem().transformToMultiAndMerge(this::createPagesFromSitemap);
    }

    private Multi<Page> createPagesFromSitemap(final GroupeDePages groupeDePages) {
        final List<String> sitemapUrls = this.readSitemap(groupeDePages.getSitemap()).getUrls().stream().toList();

        if (sitemapUrls.isEmpty()) {
            return Multi.createFrom().empty();
        } else {
            groupeDePages.setModifiedDate(LocalDate.now());
        }
        pageRepository.filterOutdatedUrls(sitemapUrls, groupeDePages);

        return Multi.createFrom().items(sitemapUrls.stream())
                .onItem().transform(value -> new PageFormDto(null, value))
                .onItem().transform(value -> this.createPageIfNotExist(value, groupeDePages));
    }

    private Page createPageIfNotExist(final PageFormDto pageFormDto, final GroupeDePages groupeDePages) {
        final Optional<Page> optionalPage = pageRepository.findByUrlAndGroupeDePages(pageFormDto.getUrl(), groupeDePages);

        Page page;
        if (optionalPage.isPresent()) {
            page = optionalPage.get();
        } else {
            page = new Page();
            page.setUrl(pageFormDto.getUrl());
            page.setGroupeDePages(groupeDePages);
            page.setAnalyseStatut(AnalyseStatut.NOUVEAU);
            page.setModifiedAt(LocalDateTime.now());
            pageRepository.persist(page);
        }
        return page;
    }

    protected boolean isValidUrl(final String url) {
        try {
            new URI(url).toURL();
            return true;
        } catch (IllegalArgumentException | URISyntaxException | IOException e) {
            return false;
        }
    }

}
