package fr.numeco.misis.sitemap.service.impl;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.time.Duration;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.ParserConfigurationException;

import org.junit.jupiter.api.Assertions;
import static org.junit.jupiter.api.Assertions.assertThrows;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyInt;
import static org.mockito.ArgumentMatchers.anyList;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.ArgumentMatchers.eq;
import org.mockito.Mock;
import org.mockito.Mockito;
import static org.mockito.Mockito.mock;
import org.mockito.MockitoAnnotations;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import fr.numeco.misis.enums.SitemapMessage;
import fr.numeco.misis.sitemap.dto.SitemapDto;
import fr.numeco.misis.sitemap.exception.SitemapUrlException;
import fr.numeco.misis.suividesite.domain.GroupeDePages;
import fr.numeco.misis.suividesite.domain.Page;
import fr.numeco.misis.suividesite.domain.SuiviDeSite;
import fr.numeco.misis.suividesite.repository.GroupeDePagesRepository;
import fr.numeco.misis.suividesite.repository.PageRepository;
import io.quarkus.test.InjectMock;
import io.quarkus.test.junit.QuarkusTest;
import io.smallrye.mutiny.helpers.test.AssertSubscriber;
import jakarta.inject.Inject;

@QuarkusTest
public class SitemapServiceUnitTest {

    @Inject
    DefaultSitemapService defaultSitemapService;

    @InjectMock
    DocumentBuilder dBuilder;

    @Mock
    Document parsedSitemap;
    @Mock
    Element element;
    @Mock
    NodeList urlNodeList;

    @Mock
    PageRepository pageRepository;
    @Mock
    GroupeDePagesRepository groupeDePagesRepository;

    @BeforeEach
    void setUp() throws IOException, SAXException {
        MockitoAnnotations.openMocks(this);

        defaultSitemapService.setPageRepository(pageRepository);
        defaultSitemapService.setGroupeDePagesRepository(groupeDePagesRepository);

        Mockito.doReturn(parsedSitemap).when(dBuilder).parse(anyString());
        Mockito.doReturn(element).when(parsedSitemap).getDocumentElement();
        Mockito.doReturn(urlNodeList).when(parsedSitemap).getElementsByTagName("loc");
    }

    private void mockSitemapResponse(List<String> nodeList) {
        Mockito.doReturn(nodeList.size()).when(urlNodeList).getLength();

        for (int i = 0; i < nodeList.size(); i++) {
            Element node = mock(Element.class);
            Mockito.doReturn(node).when(urlNodeList).item(i);
            Mockito.doReturn(Node.ELEMENT_NODE).when(node).getNodeType();
            Mockito.doReturn(nodeList.get(i)).when(node).getTextContent();

        }
    }

    private void mockDatabaseResponse(Map<String, List<String>> sitemapUrlWithPages) {
        final List<GroupeDePages> mockedGroupesDePages = new ArrayList<>();
        final SuiviDeSite suiviDeSite = mock(SuiviDeSite.class);
        Mockito.doReturn(Optional.empty()).when(pageRepository).findByUrlAndGroupeDePages(eq("https://sample.com/not-existing"), any(GroupeDePages.class));

        for (Map.Entry<String, List<String>> entry : sitemapUrlWithPages.entrySet()) {
            final GroupeDePages groupeDePages = mock(GroupeDePages.class);
            final String sitemapUrl = entry.getKey();

            Mockito.doReturn(sitemapUrl).when(groupeDePages).getSitemap();
            Mockito.doReturn(suiviDeSite).when(groupeDePages).getSuiviDeSite();

            for (String url : entry.getValue()) {
                final Page page = mock(Page.class);
                Mockito.doReturn(Optional.of(page)).when(pageRepository).findByUrlAndGroupeDePages(url, groupeDePages);
            }

            mockedGroupesDePages.add(groupeDePages);
        }

        Mockito.doReturn(mockedGroupesDePages).when(groupeDePagesRepository).findSitemapGroups();
    }

    @Test
    void shouldReturnExceptionWhenGetNotAnXMLURL() {
        String provided = "http://test.com";

        assertThrows(SitemapUrlException.class, () -> defaultSitemapService.readSitemap(provided));
    }

    @Test
    void shouldReturnExceptionWhenGetNullURL() {
        assertThrows(SitemapUrlException.class, () -> defaultSitemapService.readSitemap(null));
    }

    @Test
    void shouldReturnAListWhenGetXMLURL() throws ParserConfigurationException {
        String provided = "https://www.sample.gouv.fr/sitemap.xml";
        List<String> expected = new ArrayList<>();
        mockSitemapResponse(expected);
        Set<String> actual = defaultSitemapService.readSitemap(provided).getUrls();

        Mockito.verify(parsedSitemap, Mockito.times(1)).getElementsByTagName(anyString());
        Mockito.verify(urlNodeList, Mockito.times(1)).getLength();
        Mockito.verify(urlNodeList, Mockito.never()).item(anyInt());

        Assertions.assertIterableEquals(expected, actual);
    }

    @Test
    void shouldReturnFalseWhenInvalidUrlIsProvided() throws MalformedURLException, URISyntaxException {
        String provided = "ko://www.sample.gouv.fr";
        Boolean expected = Boolean.FALSE;

        Boolean actual = defaultSitemapService.isValidUrl(provided);

        Assertions.assertEquals(expected, actual);
    }

    @Test
    void shouldReturnTrueWhenValidUrlIsProvided() throws MalformedURLException, URISyntaxException {
        String provided = "http://www.sample.gouv.fr";
        Boolean expected = Boolean.TRUE;

        Boolean actual = defaultSitemapService.isValidUrl(provided);

        Assertions.assertEquals(expected, actual);
    }

    @Test
    void shouldReturnEmptyListWhenGetXMLURL() {
        List<String> sitemapData = List.of("ko://www.sample.fr");
        mockSitemapResponse(sitemapData);

        String provided = "https://www.sample.gouv.fr/sitemap.xml";
        List<String> expected = List.of();
        Set<String> actual = defaultSitemapService.readSitemap(provided).getUrls();

        Mockito.verify(parsedSitemap, Mockito.times(1)).getElementsByTagName(anyString());
        Mockito.verify(urlNodeList, Mockito.times(1)).getLength();
        Mockito.verify(urlNodeList, Mockito.times(1)).item(anyInt());

        Assertions.assertIterableEquals(expected, actual);
    }

    @Test
    void shouldReturnAListOfURLsWhenGetXMLURL() {
        String provided = "https://www.sample.gouv.fr/sitemap.xml";
        List<String> expected = List.of("https://www.sample.gouv.fr");
        mockSitemapResponse(expected);
        Set<String> actual = defaultSitemapService.readSitemap(provided).getUrls();

        Mockito.verify(parsedSitemap, Mockito.times(1)).getElementsByTagName(anyString());
        Mockito.verify(urlNodeList, Mockito.times(1)).getLength();
        Mockito.verify(urlNodeList, Mockito.times(1)).item(anyInt());

        Assertions.assertIterableEquals(expected, actual);
    }

    @Test
    void shouldReturnListWithUniqueUrlWhenSitemapHasDuplicated() {
        List<String> sitemapData = List.of(
                "https://www.sample-1.fr",
                "https://www.sample.fr/labas",
                "https://www.sample.fr",
                "https://www.sample.fr"
        );
        mockSitemapResponse(sitemapData);

        String provided = "https://www.sample.gouv.fr/sitemap.xml";
        List<String> expected = List.of(
                "https://www.sample-1.fr",
                "https://www.sample.fr/labas",
                "https://www.sample.fr"
        );
        Set<String> actual = defaultSitemapService.readSitemap(provided).getUrls();

        Assertions.assertIterableEquals(expected, actual);
    }

    @Test
    void shouldReturnLessUrlsWhenConfigIsProvided() {
        List<String> sitemapData = List.of(
                "https://www.sample-1.fr",
                "https://www.sample.fr/labas",
                "https://www.sample.fr/je-flemme",
                "https://www.sample.fr",
                "https://www.sample.fr"
        );
        mockSitemapResponse(sitemapData);

        String provided = "https://www.sample.gouv.fr/sitemap.xml";
        List<String> expected = List.of(
                "https://www.sample-1.fr",
                "https://www.sample.fr/labas",
                "https://www.sample.fr/je-flemme"
        );
        SitemapDto actual = defaultSitemapService.readSitemap(provided);

        Assertions.assertIterableEquals(expected, actual.getUrls());
        Assertions.assertTrue(actual.getMessages().stream().anyMatch(SitemapMessage.NUMBER_OF_ITEMS_LIMIT_REACHED::equals));
    }

    @Test
    void shouldCallOutdatedUrlsFilter() {
        List<String> sitemapData = List.of("https://sample.com/");
        mockSitemapResponse(sitemapData);
        mockDatabaseResponse(Map.of("https://sample.com/sitemap.xml", sitemapData));

        final AssertSubscriber<Page> subscriber = defaultSitemapService.process()
                .subscribe().withSubscriber(AssertSubscriber.create(1));

        subscriber
                .awaitItems(1, Duration.ofMillis(500))
                .assertCompleted();

        Mockito.verify(pageRepository, Mockito.times(1)).filterOutdatedUrls(anyList(), any());
    }

    @Test
    void shouldPersistNewPageWhenIsNotPresentOnProvidedSitemap() {
        List<String> sitemapData = List.of("https://sample.com/existing-1","https://sample.com/existing-2", "https://sample.com/not-existing");
        mockSitemapResponse(sitemapData);

        mockDatabaseResponse(Map.of("https://sample.com/sitemap.xml", List.of("https://sample.com/existing-1", "https://sample.com/existing-2")));

        final AssertSubscriber<Page> subscriber = defaultSitemapService.process()
                .subscribe().withSubscriber(AssertSubscriber.create(sitemapData.size()));

        subscriber
                .awaitItems(3, Duration.ofMillis(500))
                .assertCompleted();

        Mockito.verify(pageRepository, Mockito.times(1)).persist(any(Page.class));
    }

    @Test
    void shouldReturnEmptyMessageAlertWhenSitemapIsEmpty() {
        List<String> sitemapData = List.of();
        mockSitemapResponse(sitemapData);

        String provided = "https://www.sample.gouv.fr/sitemap.xml";
        List<String> expected = List.of();
        SitemapDto actual = defaultSitemapService.readSitemap(provided);

        Assertions.assertIterableEquals(expected, actual.getUrls());
        Assertions.assertTrue(actual.getMessages().stream().anyMatch(SitemapMessage.EMPTY_SITEMAP::equals));
    }

}
